{{
Original Tiny Basic written by Tomas Rokicki & Radical Eye Software.

Copyright (c) 2008,2020 Michael Green.  See end of file for terms of use.
}}

con
  _XTALFREQ     = 20_000_000                                    ' crystal frequency
  _XDIV         = 4             '\                              '\ crystal divider                      to give 5.0MHz
  _XMUL         = 72            '| 180MHz                       '| crystal / div * mul                  to give 360MHz
  _XDIVP        = 2             '/                              '/ crystal / div * mul /divp            to give 180MHz
  _XOSC         = %01                                   'OSC    ' %00=OFF, %01=OSC, %10=15pF, %11=30pF
  _XSEL         = %11                                   'XI+PLL ' %00=rcfast(20+MHz), %01=rcslow(~20KHz), %10=XI(5ms), %11=XI+PLL(10ms)
  _XPPPP        = ((_XDIVP>>1) + 15) & $F                       ' 1->15, 2->0, 4->1, 6->2...30->14
  _CLOCKFREQ    = _XTALFREQ / _XDIV * _XMUL / _XDIVP            ' internal clock frequency                
  _SETFREQ      = 1<<24 + (_XDIV-1)<<18 + (_XMUL-1)<<8 + _XPPPP<<4 + _XOSC<<2  ' %0000_000e_dddddd_mmmmmmmmmm_pppp_cc_00  ' setup  oscillator
  _ENAFREQ      = _SETFREQ + _XSEL                                             ' %0000_000e_dddddd_mmmmmmmmmm_pppp_cc_ss  ' enable oscillator

  _HUBROM       = $FC000               ' ROM $FC000
  _HUBBUF       = $FC000               ' overwrite Booter
  _HUBBUFSIZE   = 80                   ' RxString default size for _HUBBUF

  _baud         = 115_200
  _bitper       = (_clockfreq / _baud) << 16 + 7          ' 115200 baud, 8 bits
  _txmode       = %0000_0000_000_0000000000000_01_11110_0 'async tx mode, output enabled for smart output
  _rxmode       = %0000_0000_000_0000000000000_00_11111_0 'async rx mode, input  enabled for smart input
  _rx_pin       = 63                   ' pin serial receiver
  _tx_pin       = 62                   ' pin serial transmitter

   version   = 1                       ' Major version
   release   = 1                       ' Minor release
   testLevel = 6                       ' Test change level
   
   progsize  = 8192                    ' Space reserved for program
   maxstack  = 20                      ' Maximum stack depth
   linelen   = 256                     ' Maximum input line length
   quote     = 34                      ' Double quote
   caseBit   = !32                     ' Uppercase/Lowercase bit

'------[ LMM parameters, etc ]--------------------------------------------------------------
lmm_x        = $1e0 'parameter passed to/from LMM routine (typically a value)
lmm_f        = $1e1 'parameter passed to      LMM routine (function options; returns unchanged)
lmm_p        = $1e2 'parameter passed to/from LMM routine (typically a hub/cog ptr/addr)
lmm_p2       = $1e3 'parameter passed to/from LMM routine (typically a 2nd hub/cog address)
lmm_c        = $1e4 'parameter passed to/from LMM routine (typically a count)

'-------[ LMM additional workareas ]-------------------------------------------
lmm_w        = $1e5 'workarea (never saved - short term use between calls, except _HubTx)
lmm_tx       = $1e6 '_HubTx
lmm_hx       = $1e7 '_HubHex/_HubString
lmm_hx2      = $1e8 '_HubHex
lmm_hc       = $1e9 '  "
lmm_lx       = $1ea '_HubList
lmm_lf       = $1eb '  "
lmm_lp       = $1ec '  "
lmm_lp2      = $1ed '  "
lmm_lc       = $1ee '  "
lmm_bufad    = $1ef '_HubRxString

'-------[ LMM ASCII equates ]-------------------------------------------
  _CLS_      = $0C     
  _BS_       = $08
  _LF_       = $0A
  _CR_       = $0D
  _ESC_      = $1B
  _DEL_      = $7F
  _EOF_      = $FF

'-------[ lmm_f CALL Modes...(b4..b0 are modifier options) ]-------------------------------------------
  _MODE         = $F << 5                       ' mode bits defining the call b8..b5 (b4..b0 are modifier options)
  _SHIFT        = 5                             ' shr # to extract mode bits
  _HEX_         = 2 << 5                        ' hex...
    _REV_               = 1 << 4                '   - reverse byte order
    _SP                 = 1 << 3                '   - space between hex output pairs
   '_DIGITS             = 7..0 where 8->0       '   - no. of digits to display
  _LIST         = 3 << 5                        ' LIST memory line (1/4 longs) from cog/hub
    _ADDR2              = 1 << 4                ' 1= use lmm_p2 as to-address
    _LONG_              = 1 << 1                ' 1=display longs xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx
  _TXSTRING     = 4 << 5                        ' tx string (nul terminated) from hub
  _RXSTRING     = 5 << 5                        ' rx string
    _ECHO_              = 1 << 4                '    - echo char
    _PROMPT             = 1 << 3                '    - prompt (lmm_x)
    _ADDR               = 1 << 2                '    - addr of string buffer supplied
    _NOLF               = 1 << 1                '    - strip <lf>
  _MONITOR      = 7 << 5                        ' goto rom monitor

'-------[ LMM routines callable from users' programs ]-------------------------------------------
  _SerialInit      = $fcab8     ' Serial Initialise     (lmm_x & lmm_bufad must be set first)
  _HubTxCR         = $fcae4     ' Sends <cr><lf>        (overwrites lmm_x)
  _HubTxRev        = $fcaec     ' Sends lmm_x with bytes reversed
  _HubTx           = $fcaf0     ' Sends lmm_x           (can be up to 4 bytes)
  _HubHexRev       = $fcb24     ' Sends lmm_x with bytes reversed as Hex char(s) as defined in lmm_f
  _HubHex8         = $fcb28     ' Sends lmm_x as Hex char(s) after setting lmm_f as 8 hex chars
  _HubHex          = $fcb2c     ' Sends lmm_x as Hex char(s) as defined in lmm_f
  _HubTxStrVer     = $fcb9c     ' Sends $0 terminated string at lmm_p address after setting lmm_p=##_str_vers
  _HubTxString     = $fcba4     ' Sends $0 terminated string at lmm_p address
  _HubListA2H      = $fcbc4     ' List/Dump line(s) from lmm_p address to lmm_p2 address after setting lmm_f=#_LIST+_ADDR2 
  _HubList         = $fcbc8     ' List/Dump line(s) from lmm_p address to lmm_p2 address according to lmm_f
  _HubRx           = $fcb10     ' Recv char into lmm_x
  _HubRxStrMon     = $fccc4     ' Recv string into lmm_bufad address after setting prompt=lmm_x=#"*" & params=lmm_f=#_RXSTRING+_ECHO_+_PROMPT
  _HubRxString     = $fcccc     ' Recv string into lmm_p/lmm_bufad address according to params in lmm_f
  _HubMonitor      = $fcd78     ' Calls the Monitor; uses lmm_bufad as the input buffer address
  _RdLongCogHub    = $fcf34     ' read cog/lut/hub long from lmm_p address into lmm_x, then lmm_p++
  _str_vers        = $fd014     ' locn of hub string, $0 terminated

var
   long sp, tp, eop, nextlineloc, rv, curlineno, pauseTime
   long vars[26], stack[maxstack], fcb[8], outset[2]
   long forStep[26], forLimit[26], forLoop[26]
   long outputs[2], progdata[progsize/4]
   byte tline[linelen], tailLine[linelen], inVars[26]

obj
   ser : "SmartSerial"

con
   #128, tok_if, tok_then, tok_input, tok_print, tok_goto, tok_gosub, tok_return, tok_rem
   tok_new, tok_list, tok_run, tok_rnd, tok_open, tok_read, tok_write, tok_close
   tok_delete, tok_rename, tok_files, tok_save, tok_load, tok_not, tok_and, tok_or
   tok_shl, tok_shr, tok_for, tok_to, tok_step, tok_next, tok_pinread, tok_pinwrite
   tok_pause, tok_using, tok_rol, tok_ror, tok_sar, tok_rev, tok_byte, tok_word
   tok_long, tok_cnt, tok_phsa, tok_phsb, tok_frqa, tok_frqb, tok_ctra, tok_ctrb
   tok_display, tok_keycode, tok_let, tok_stop, tok_end, tok_eeprom, tok_file, tok_mem
   tok_spin, tok_copy, tok_dump

dat
   tok0  byte "IF", 0
   tok1  byte "THEN", 0
   tok2  byte "INPUT", 0    ' INPUT {"<prompt>";} <var> {,<var>}
   tok3  byte "PRINT", 0    ' PRINT {USING "<format>";} ...
   tok4  byte "GOTO", 0
   tok5  byte "GOSUB", 0
   tok6  byte "RETURN", 0
   tok7  byte "REM", 0
   tok8  byte "NEW", 0
   tok9  byte "LIST", 0
   tok10 byte "RUN", 0
   tok11 byte "RND", 0
   tok12 byte "OPEN", 0     ' OPEN " <file> ",<mode>
   tok13 byte "READ", 0     ' READ <var> {,<var>}
   tok14 byte "WRITE", 0    ' WRITE {USING "<format>";} ...
   tok15 byte "CLOSE", 0    ' CLOSE
   tok16 byte "DELETE", 0   ' DELETE " <file> "
   tok17 byte "RENAME", 0   ' RENAME " <file> "," <file> "
   tok18 byte "FILES", 0    ' FILES
   tok19 byte "SAVE", 0     ' SAVE or SAVE [<expr>] or SAVE "<file>"
   tok20 byte "LOAD", 0     ' LOAD or LOAD [<expr>] or LOAD "<file>"
   tok21 byte "NOT" ,0      ' NOT <logical>
   tok22 byte "AND" ,0      ' <logical> AND <logical>
   tok23 byte "OR", 0       ' <logical> OR <logical>
   tok24 byte "SHL", 0      ' <expr> SHL <expr>
   tok25 byte "SHR", 0      ' <expr> SHR <expr>
   tok26 byte "FOR", 0      ' FOR <var> = <expr> TO <expr>
   tok27 byte "TO", 0
   tok28 byte "STEP", 0     '  optional STEP <expr>
   tok29 byte "NEXT", 0     ' NEXT <var>
   tok30 byte "PINREAD", 0  ' PINREAD [ <expr> {.. <expr>} ]
   tok31 byte "PINWRITE", 0 ' PINWRITE [ <expr> {.. <expr>} ] = <expr>
   tok32 byte "PAUSE", 0    ' PAUSE <time ms> {,<time us>}
   tok33 byte "USING", 0    ' PRINT USING "<format>"; ...
   tok34 byte "ROL", 0      ' <expr> ROL <expr>
   tok35 byte "ROR", 0      ' <expr> ROR <expr>
   tok36 byte "SAR", 0      ' <expr> SAR <expr>
   tok37 byte "REV", 0      ' <expr> REV <expr>
   tok38 byte "BYTE", 0     ' BYTE [ <expr> ]
   tok39 byte "WORD", 0     ' WORD [ <expr> ]
   tok40 byte "LONG", 0     ' LONG [ <expr> ]
   tok41 byte "CNT", 0
   tok42 byte "PHSA", 0
   tok43 byte "PHSB", 0
   tok44 byte "FRQA", 0
   tok45 byte "FRQB", 0
   tok46 byte "CTRA", 0
   tok47 byte "CTRB", 0
   tok48 byte "DISPLAY", 0  ' DISPLAY <expr> {,<expr>}
   tok49 byte "KEYCODE", 0  ' KEYCODE
   tok50 byte "LET", 0
   tok51 byte "STOP", 0
   tok52 byte "END", 0
   tok53 byte "EEPROM", 0   ' EEPROM[ <expr> ]
   tok54 byte "FILE", 0     ' FILE
   tok55 byte "MEM", 0      ' MEM
   tok56 byte "SPIN", 0     ' SPIN [<expr>] or SPIN "<file>"
   tok57 byte "COPY", 0     ' COPY [<expr>],"<file>" or COPY "<file>",[<expr>] or
                            ' COPY [<expr>],<expr> where <expr> are different
   tok58 byte "DUMP", 0     ' DUMP <expr>,<expr> or DUMP [<expr>],<expr>

   toks  word @tok0,  @tok1,  @tok2,  @tok3,  @tok4,  @tok5,  @tok6,  @tok7
         word @tok8,  @tok9,  @tok10, @tok11, @tok12, @tok13, @tok14, @tok15
         word @tok16, @tok17, @tok18, @tok19, @tok20, @tok21, @tok22, @tok23
         word @tok24, @tok25, @tok26, @tok27, @tok28, @tok29, @tok30, @tok31
         word @tok32, @tok33, @tok34, @tok35, @tok36, @tok37, @tok38, @tok39
         word @tok40, @tok41, @tok42, @tok43, @tok44, @tok45, @tok46, @tok47
         word @tok48, @tok49, @tok50, @tok51, @tok52, @tok53, @tok54, @tok55
         word @tok56, @tok57, @tok58
   tokx  word

   syn   byte "Syntax Error", 0
   ln    byte "Invalid Line Number", 0

PUB main | err, s, i
   ser.Start(_rx_pin,_tx_pin,%0000,_baud)
   consoleStr(string("FemtoDongle2"))
   if version > 0 or release > 0 or testLevel > 0
     consoleStr(string(" Version "))
     consoleDec(version)
     consoleOut(".")
     if release < 100
       consoleOut("0")
     if release < 10
       consoleOut("0")
     consoleDec(release)
     if testLevel > 0
       consoleOut("a"+testLevel-1)
   consoleOut(_CR_)
'' Clear the program space and variables, then read a line and interpret it.
   pauseTime := 0
   longfill(@fcb,0,8)
   outset[0] := outputs[0] := %00000000_00000000_00000000_00000000
   outset[1] := outputs[1] := %11000000_00000000_00000000_00000000
   clearall
   s := 0
   curlineno := -1
   repeat
      err := \doline(s)
      s := 0
      if err
         showError(err)

PRI showError(err)
   if curlineno => 0
      consoleStr(string("IN LINE "))
      consoleDec(curlineno)
      consoleOut(" ")
   if err < 0
      consoleStr(string("SD card "))
      consoleDec(err)
      consoleOut(_CR_)
   else
      putlinet(err)
   nextlineloc := eop - 2

pri consoleIn                        ' Get keyboard character from proper device
   result := ser.rxcheck

pri consoleOut(c)                    ' Display on proper device
   ser.tx(c)
   if c == _CR_                      ' Convert CR to CR/LF
      ser.tx(_LF_)

PUB consoleStr(s)                    ' Display string at s
   repeat strsize(s)
      consoleOut(byte[s++])

PUB consoleDec(v) | i, x             ' Display decimal number
   x := v == NEGX                    ' Check for max negative
   if v < 0
      v := ||(v+x)                   ' Adjust as needed
      consoleOut("-")                '  and output sign
   i := 1_000_000_000                ' Initialize divisor
   repeat 10                         ' Loop for 10 digits
      if v => i
         consoleOut(v / i + "0" + x*(i == 1)) 'display digit
         v //= i
         result~~                    ' Flag non-zero found
      elseif result or i == 1
         consoleOut("0")             ' Display at least zero
      i /= 10

PUB consoleHex(v,w)                  ' Display hexadecimal
   v <<= (8 - w) << 2                '  at most 8 digits
   repeat w
      consoleOut(lookupz((v <-= 4) & $F : "0".."9", "A".."F"))

PUB consoleBin(v, w)                 ' Display binary
   v <<= 32 - w                      '  at most 32 digits
   repeat w
      consoleOut((v <-= 1) & 1 + "0")

pri massIn
'  result is a byte from the currently open file
   return 0

pri massOut(c)                       ' Output to mass storage device
'  write c to the currently open file
   return 0

pri massStr(s) | c                   ' Output to mass storage device
   repeat while c := byte[s++]
      massOut(c)

pri massDec(v) | i, x                ' Output to mass storage device
   x := v == NEGX                    ' Check for max negative
   if v < 0
      v := ||(v+x)                   ' Adjust as needed
      massOut("-")                   '  and output sign
   i := 1_000_000_000                ' Initialize divisor
   repeat 10                         ' Loop for 10 digits
      if v => i
         massOut(v / i + "0" + x*(i == 1)) 'display digit
         v //= i
         result~~                    ' Flag non-zero found
      elseif result or i == 1
         massOut("0")                ' Display at least zero
      i /= 10

pri massOpen
   return 0

pri massCreate
   return 0

pri massClose
   return 0

pri massBoot
   return 0

PRI getline | i, c
   i := 0
   repeat
      repeat while (c := consoleIn) < 0
      if (c == _BS_) or (c == _DEL_)
         if i > 0
            consoleStr(string(_BS_," ",_BS_))
            i--
      elseif c == _CR_
         consoleOut(_CR_)
         tline[i] := 0
         tp := @tline
         return
      elseif i < linelen-1
         consoleOut(c)
         tline[i++] := c

pri putlinet(s) | c, ntoks
   ntoks := (@tokx - @toks) / 2
   repeat while c := byte[s++]
      if c => 128
         if (c -= 128) < ntoks
            consoleStr(@@toks[c])
            if c <> tok_rem   ' REM
               consoleOut(" ")
         else
            consoleOut("{")
            consoleDec(c)
            consoleOut("}")
      else
         consoleOut(c)
   consoleOut(_CR_)

pri spaces | c
   repeat
      c := byte[tp]
      if c == 0 or c > " "
         return c
      tp++

pri skipspaces
   if byte[tp]
      tp++
   return spaces

pri parseliteral | r, c
   r := 0
   repeat
      c := byte[tp]
      if c < "0" or c > "9"
         return r
      r := r * 10 + c - "0"
      tp++

pri movprog(at, delta)
   if eop + delta + 2 - @progdata > progsize
      abort string("NO MEMORY")
   bytemove(at+delta, at, eop-at)
   eop += delta

pri fixvar(c)
   if c => "a"
      c -= 32
   return c - "A"

pri isvar(c)
   c := fixvar(c)
   return c => 0 and c < 26

pri tokenize | tok, c, at, put, state, i, j, ntoks
   ntoks := (@tokx - @toks) / 2
   at := tp
   put := tp
   state := 0
   repeat while c := byte[at]
      if c == quote
         if state == "Q"
            state := 0
         elseif state == 0
            state := "Q"
      if state == 0
         repeat i from 0 to ntoks-1
            tok := @@toks[i]
            j := 0
            repeat while byte[tok] and ((byte[tok] ^ byte[j+at]) & caseBit) == 0
               j++
               tok++
            if byte[tok] == 0 and not isvar(byte[j+at])
               byte[put++] := 128 + i
               at += j
               if i == (tok_rem - 128)
                  state := "R"
               else
                  repeat while byte[at] == " "
                     at++
                  state := "F"
               quit
         if state == "F"
            state := 0
         else
            byte[put++] := byte[at++]
      else
         byte[put++] := byte[at++]
   byte[put] := 0

pri wordat(loc)
   return (byte[loc]<<8)+byte[loc+1]

pri findline(lineno) | at
   at := @progdata
   repeat while wordat(at) < lineno
      at += 3 + strsize(at+2)
   return at

pri insertline | lineno, fc, loc, locat, newlen, oldlen
   lineno := parseliteral
   if lineno < 0 or lineno => 65535
      abort @ln
   tokenize
   fc := spaces
   loc := findline(lineno)
   locat := wordat(loc)
   newlen := 3 + strsize(tp)
   if locat == lineno
      oldlen := 3 + strsize(loc+2)
      if fc == 0
         movprog(loc+oldlen, -oldlen)
      else
         movprog(loc+oldlen, newlen-oldlen)
   elseif fc
      movprog(loc, newlen)
   if fc
      byte[loc] := lineno >> 8
      byte[loc+1] := lineno
      bytemove(loc+2, tp, newlen-2)

pri clearvars
   longfill(@vars, 0, 26)
   pauseTime := 0
   nextlineloc := @progdata
   sp := 0

pri newprog
   nextlineloc := @progdata
   progdata[0] := $0000_FFFF
   eop := @progdata + 2
   sp := 0

pri clearall
   newprog
   clearvars

pri pushstack
   if sp => constant(maxstack-1)
      abort string("RECURSION ERROR")
   stack[sp++] := nextlineloc

pri getAddress(delim) | t
   if spaces <> "["
      abort @syn
   skipspaces
   result := expr
   if delim == "." and (result < 0 or result > 63)
      abort string("Invalid pin number")
   if delim == "." or delim == ","
      if spaces == delim
         if delim == "."             ' Handle the form <expr>..<expr>
            if byte[++tp] <> "."
               abort @syn
            result <<= 8
            skipspaces
            t := expr
            if t < 0 or t > 63
               abort string("Invalid pin number")
            result |= t | $10000
         else                        ' Handle the form <expr>,<expr>
            if result & 1 or result < 0 or result > 63
               abort string("Invalid pin number")
            skipspaces
            result := (result << 18) | (expr & $7FFFF)
      elseif delim == ","
         result := (result & $7FFFF) { | mem#bootAddr }
   if spaces <> "]"
      abort @syn
   tp++

pri factor | tok, t, i, a
   tok := spaces
   tp++
   case tok
      "(":
         t := expr
         if spaces <> ")"
            abort @syn
         tp++
         return t
      "a".."z","A".."Z":
         return vars[fixvar(tok)]
      tok_pinread: ' PINREAD [ <expr>{..<expr>} ]
         t := getAddress(".")
         if t > $FFFF
           tok := t & $FF
           t := (t >> 8) & $FF
           abort string("Multiple pins not implemented yet")
         else
           outputs[t>>5] &= ! |< (t & $1F)
           return pinread(t)
      tok_byte: ' BYTE [ <expr> ]
         return byte[getAddress(" ")]
      tok_word: ' WORD [ <expr> ]
         return word[getAddress(" ")]
      tok_long: ' LONG [ <expr> ]
         return long[getAddress(" ")]
      tok_cnt: ' CNT
         return GETCT()
      tok_phsa: ' PHSA
         abort string("Not available on P2")
      tok_phsb: ' PHSB
         abort string("Not available on P2")
      tok_frqa: ' FRQA
         abort string("Not available on P2")
      tok_frqb: ' FRQB
         abort string("Not available on P2")
      tok_ctra: ' CTRA
         abort string("Not available on P2")
      tok_ctrb: ' CTRB
         abort string("Not available on P2")
      tok_keycode: ' KEYCODE
         return consoleIn
      tok_rnd: ' RND <factor>
         return (rv? >> 1) ** (factor << 1)
      "-":
         return - factor
      "!":
         return ! factor
      "$", "%", quote, "0".."9":
         --tp
         return getAnyNumber
      other:
         abort(@syn)

pri shifts | tok, t
   t := factor
   tok := spaces
   if tok == tok_shl ' SHL
      tp++
      return t << factor
   elseif tok == tok_shr ' SHR
      tp++
      return t >> factor
   elseif tok == tok_rol ' ROL
      tp++
      return t <- factor
   elseif tok == tok_ror ' ROR
      tp++
      return t -> factor
   elseif tok == tok_sar ' SAR
      tp++
      return t ~> factor
   elseif tok == tok_rev ' REV
      tp++
      return t >< factor
   else
      return t

pri bitFactor | tok, t
   t := shifts
   repeat
      tok := spaces
      if tok == "&"
         tp++
         t &= shifts
      else
         return t

pri bitTerm | tok, t
   t := bitFactor
   repeat
      tok := spaces
      if tok == "|"
         tp++
         t |= bitFactor
      elseif tok == "^"
         tp++
         t ^= bitFactor
      else
         return t

pri term | tok, t
   t := bitTerm
   repeat
      tok := spaces
     if tok == "*"
        tp++
        t *= bitTerm
     elseif tok == "/"
        if byte[++tp] == "/"
           tp++
           t //= bitTerm
        else
           t / =bitTerm
     else
        return t

pri arithExpr | tok, t
   t := term
   repeat
      tok := spaces
      if tok == "+"
         tp++
         t += term
      elseif tok == "-"
         tp++
         t -= term
      else
         return t

pri compare | op, a, b, c
   a := arithExpr
   op := 0
   spaces
   repeat
      c := byte[tp]
      case c
         "<": op |= 1
              tp++
         ">": op |= 2
              tp++
         "=": op |= 4
              tp++
         other: quit
   case op
      0: return a
      1: return a < arithExpr
      2: return a > arithExpr 
      3: return a <> arithExpr
      4: return a == arithExpr
      5: return a =< arithExpr
      6: return a => arithExpr
      7: abort string("Invalid comparison")

pri logicNot | tok
   tok := spaces
   if tok == tok_not ' NOT
      tp++
      return not compare
   return compare

pri logicAnd | t, tok
   t := logicNot
   repeat
      tok := spaces
      if tok == tok_and ' AND
         tp++
         t := t and logicNot
      else
         return t

pri expr | tok, t
   t := logicAnd
   repeat
      tok := spaces
      if tok == tok_or ' OR
         tp++
         t := t or logicAnd
      else
         return t

pri specialExpr
   if spaces <> "="
      abort @syn
   skipspaces
   return expr

pri scanFilename(f) | c, chars
   chars := 0
   tp++ ' skip past initial quote
   repeat while (c := byte[tp++]) <> quote
      if chars++ < 31
         byte[f++] := c
   byte[f] := 0

pri texec | ht, nt, restart, thisLine, a,b,c,d,e
   thisLine := tp - 2
   restart := 1
   repeat while restart
      restart := 0
      ht := spaces
      if ht == 0
         return
      nt := skipspaces
      if isvar(ht) and nt == "="
         tp++
         vars[fixvar(ht)] := expr
      elseif ht => 128
         case ht
            tok_if: ' IF
               a := expr
               if spaces <> 129
                  abort string("MISSING THEN")
               skipspaces
               if not a
                  return
               restart := 1
            tok_input: ' INPUT {"<prompt>";} <var> {, <var>}
               if nt == quote
                  c := byte[++tp]
                  repeat while c <> quote and c
                     consoleOut(c)
                     c := byte[++tp]
                  if c <> quote
                     abort @syn
                  if skipspaces <> ";"
                     abort @syn
                  nt := skipspaces
               if not isvar(nt)
                  abort @syn
               b := 0
               inVars[b++] := fixvar(nt)
               repeat while skipspaces == ","
                  nt := skipspaces
                  if not isvar(nt) or b == 26
                     abort @syn
                  inVars[b++] := fixvar(nt)
               getline
               tokenize
               repeat a from 1 to b
                  vars[inVars[a-1]] := expr
                  if a < b
                     if spaces == ","
                        skipspaces
            tok_print: ' PRINT
               a := 0
               repeat
                  nt := spaces
                  if nt == 0 or nt == ":"
                     quit
                  if nt == quote
                     tp++
                     repeat
                        c := byte[tp++]
                        if c == 0 or c == quote
                           quit
                        consoleOut(c)
                        a++
                  else
                     d~
                     b := expr
                     consoleDec(b)
                  nt := spaces
                  if nt == ";"
                     tp++
                  elseif nt == ","
                     consoleOut(" ")
                     a++
                     repeat while a & 7
                        consoleOut(" ")
                        a++
                     tp++
                  elseif nt == 0 or nt == ":"
                     consoleOut(_CR_)
                     quit
                  else
                     abort @syn
            tok_goto, tok_gosub: ' GOTO, GOSUB
               a := expr
               if a < 0 or a => 65535
                  abort @ln
               b := findline(a)
               if wordat(b) <> a
                  abort @ln
               if ht == tok_gosub
                  pushstack
               nextlineloc := b 
            tok_return: ' RETURN
               if sp == 0
                  abort string("INVALID RETURN")
               nextlineloc := stack[--sp]
            tok_rem: ' REM
               repeat while skipspaces
            tok_new: ' NEW
               clearall
            tok_list: ' LIST {<expr> {,<expr>}}
               b := 0                ' Default line range
               c := 65535
               if spaces <> 0        ' At least one parameter
                  b := c := expr
                  if spaces == ","
                     skipspaces
                     c := expr
               a := @progdata
               repeat while a+2 < eop
                  d := wordat(a)
                  if d => b and d =< c
                     consoleDec(d)
                     consoleOut(" ")
                     putlinet(a+2)
                  a += 3 + strsize(a+2)
            tok_run: ' RUN
                  clearvars
            tok_open: ' OPEN " <file> ", R/W/A or OPEN [<expr>{,<expr>}]
               if spaces <> quote
                  abort @syn
               scanFilename(@fcb)
               if spaces <> ","
                  abort @syn
               case skipspaces
                  "A", "a": d := "a"
                  "W", "w": d := "w"
                  "R", "r": d := "r"
                  "D", "d": d := "d"
                  "X", "x": d := "x"
                  other: abort string("Invalid open file mode")
               tp++
               massClose           ' Close any open file
               ifnot massOpen      ' Try first to open file
                  if d == "w" or d == "a" ' Only create new file
                  if massCreate    '  if writing or appending
                     abort string("Can't create file")
            tok_read: ' READ <var> {, <var> }
               if not isvar(nt)
                  abort @syn
               d := 0
               inVars[d++] := fixvar(nt)
               repeat while skipspaces == ","
                  nt := skipspaces
                  if not isvar(nt) or d == 26
                     abort @syn
                  inVars[d++] := fixvar(nt)
               ifnot fcb[7] >> 24   ' No file open
                  abort string("Can't read file")
               a := 0
               repeat
                  if (c := massIn) < 0
                     abort string("Can't read file")
                  elseif c == _CR_ or c == _EOF_
                     tline[a] := 0
                     tp := @tline
                     quit
                  elseif c == _LF_
                     next
                  elseif a < linelen-1
                     tline[a++] := c
               tokenize
               repeat a from 1 to d
                  vars[inVars[a-1]] := expr
                  if a < d
                     if spaces == ","
                        skipspaces
            tok_write: ' WRITE ...
               d := 0 ' record column
               ifnot fcb[7] >> 24   ' No file open
                  abort string("Can't write to file")
               repeat
                  nt := spaces
                  if nt == 0 or nt == ":"
                     quit
                  if nt == quote
                     tp++
                     repeat
                        c := byte[tp++]
                        if c == 0 or c == quote
                           quit
                        massOut(c)
                        d++
                  else
                     a := expr
                     massDec(a)
                  nt := spaces
                  if nt == ";"
                     tp++
                  elseif nt == ","
                     massOut(" ")
                     d++
                     repeat while d & 7
                        massOut(" ")
                        d++
                     tp++
                  elseif nt == 0 or nt == ":"
                     massOut(_CR_)
                     massOut(_LF_)
                     quit
                  else
                     abort @syn
            tok_close: ' CLOSE
                  ifnot massClose
                     abort string("Error closing file")
            tok_delete: ' DELETE " <file> "
               if spaces <> quote
                  abort @syn
               scanFilename(@fcb)
               abort string("Delete not implemented")
            tok_rename: ' RENAME " <file> "," <file> "
               if spaces <> quote
                  abort @syn
               scanFilename(@fcb)
               if spaces <> ","
                  abort @syn
               if skipspaces <> quote
                  abort @syn
               scanFilename(@fcb)
               abort string("Rename not implemented")
            tok_files: ' FILES
               abort string("Files not implemented")
{               ifnot \mem.mount(sdspiDO,sdspiClk,sdspiDI,sdspiCS) < 0
                  consoleStr(string("SD Card Root Directory",_CR_))
                  mem.opendir
                  b := 0
                  d := false
                  repeat while mem.nextfile(@f0) == 0
                     d := true
                     if b == 39
                        consoleOut(_CR_)
                        waitcnt(clkfreq/20 + cnt)
                        b := 0
                     consoleStr(@f0)
                     repeat 13 - strsize(@f0)
                        consoleOut(" ")
                     b := b + 13
                  if d
                     consoleOut(_CR_)
                     waitcnt(clkfreq/20 + cnt)
                  \mem.unmount }
            tok_save: ' SAVE or SAVE "<filename>"
               if (nt := spaces) == quote
                  scanFilename(@fcb)
                  massClose           ' Close any open file
                  ifnot massOpen      ' Try first to open file
                     if d == "w" or d == "a" ' Only create new file
                        if massCreate    '  if writing or appending
                           abort string("Can't create file")
                  processSave
            tok_load: ' LOAD or LOAD "<filename>"
               if (nt := spaces) == quote
                  scanFilename(@fcb)
                  massClose           ' Close any open file
                  ifnot massOpen      ' Try first to open file
                     abort string("Can't open file")
               a := @tailLine                 ' Save statement tail
               repeat while byte[a++] := byte[tp++]
               newprog
               processLoad
               massClose
               tp := @tailLine                ' Scan copy after load
               eop := @progdata + 2
               nextlineloc := eop - 2         ' Leave it stopped
            tok_for: ' FOR <var> = <expr> TO <expr> {STEP <expr>}
               ht := spaces
               if ht == 0
                  abort @syn
               nt := skipspaces
               if not isvar(ht) or nt <> "="
                  abort @syn
               a := fixvar(ht)
               skipspaces
               vars[a] := expr
               if spaces <> tok_to ' TO          ' Save FOR limit
                  abort @syn
               skipspaces
               forLimit[a] := expr
               if spaces == tok_step ' STEP      ' Save step size
                  skipspaces
                  forStep[a] := expr
               else
                  forStep[a] := 1                ' Default step is 1
               if spaces
                  abort @syn
               forLoop[a] := nextlineloc         ' Save address of line
               if forStep[a] < 0                 '  following the FOR
                  b := vars[a] => forLimit[a]
               else                              ' Initially past the limit?
                  b := vars[a] =< forLimit[a]
               if not b                          ' Search for matching NEXT 
                  repeat while nextlineloc < eop-2
                     curlineno := wordat(nextlineloc)
                     tp := nextlineloc + 2
                     nextlineloc := tp + strsize(tp) + 1
                     if spaces == tok_next       ' NEXT <var>
                        nt := skipspaces         ' Variable has to agree
                        if not isvar(nt)
                           abort @syn
                        if fixvar(nt) == a       ' If match, continue after
                           quit                  '  the matching NEXT
            tok_next: ' NEXT <var>
               nt := spaces
               if not isvar(nt)
                  abort @syn
               a := fixvar(nt)
               vars[a] += forStep[a]             ' Increment or decrement the
               if forStep[a] < 0                 '  FOR variable and check for
                  b := vars[a] => forLimit[a]
               else                              '  the limit value
                  b := vars[a] =< forLimit[a]
               if b                              ' If continuing loop, go to
                  nextlineloc := forLoop[a]      '  statement after FOR
               tp++
            tok_pinwrite: ' PINWRITE [ <expr>{..<expr>} ] = <expr>
               a := getAddress(".")
               if a > $FFFF
                  b := a & $FF
                  a := (a >> 8) & $FF
                  c := specialExpr
                  abort string("Multiple pins not implemented yet")
               else
                  pinwrite(a,specialExpr)
                  outputs[a>>5] |= |< (a & $1F)
            tok_pause: ' PAUSE <expr> {,<expr>}
               if pauseTime == 0                 ' If no active pause time, set it
                  spaces                         '  with a minimum time of 50us
                  pauseTime := expr * 1000
                  if spaces == ","               ' First (or only) value is in ms
                     skipspaces
                     pauseTime += expr           ' Second value is in us
                  pauseTime #>= 50
               if pauseTime < 10_050             ' Normally pause at most 10ms at a time,
                  WAITUS(pauseTime)              '  but, if that would leave < 50us,
                  pauseTime := 0                 '   pause the whole amount now
               else                             
                  a := pauseTime <# 10_000
                  WAITUS(a)                      ' Otherwise, pause at most 10ms and
                  nextlineloc := thisLine        '  re-execute the PAUSE for the rest
                  pauseTime -= 10_000
            tok_byte: ' BYTE [ <expr> ] = <expr>
               a := getAddress(" ")
               byte[a] := specialExpr
            tok_word: ' WORD [ <expr> ] = <expr>
               a := getAddress(" ")
               word[a] := specialExpr
            tok_long: ' LONG [ <expr> ] = <expr>
               a := getAddress(" ")
               long[a] := specialExpr
            tok_phsa: ' PHSA =
               abort string("Not available on P2")
            tok_phsb: ' PHSB =
               abort string("Not available on P2")
            tok_frqa: ' FRQA =
               abort string("Not available on P2")
            tok_frqb: ' FRQB =
               abort string("Not available on P2")
            tok_ctra: ' CTRA =
               abort string("Not available on P2")
            tok_ctrb: ' CTRB =
               abort string("Not available on P2")
            tok_display: ' DISPLAY <expr> {,<expr>}
               spaces
               consoleOut(expr)
               repeat while spaces == ","
                  skipspaces
                  consoleOut(expr)
            tok_let: ' LET <var> = <expr>
               nt := spaces
               if not isvar(nt)
                  abort @syn
               tp++
               vars[fixvar(nt)] := specialExpr
            tok_stop: ' STOP
               nextlineloc := eop-2
               return
            tok_end: ' END
               nextlineloc := eop-2
               return
            tok_spin: ' SPIN "<file>"
               if spaces == quote
                  scanFilename(@fcb)
                  massBoot
               abort string("SPIN unsuccessful")
            tok_copy: ' COPY [<expr>],"<file>" or COPY "<file>",[<expr>] or
                      ' COPY [<expr>],[<expr>] where <expr> are different
               abort string("Copy not implemented")
            tok_dump: ' DUMP <expr>,<expr> or DUMP [<expr>],<expr>
               if spaces == "["
                  c := getAddress(",")
                  a := c & $F80000
                  b := c & $07FFFF
               else
                  a := -1
                  b := expr
               if spaces <> ","
                  abort @syn
               skipspaces
               dumpMemory(a,b,expr)
      else
         abort(@syn)
      if spaces == ":"
         restart := 1
         tp++

pri doline(s) | c                 ' Execute the string in s or wait for input
   curlineno := -1
   if nextlineloc < eop-2
      curlineno := wordat(nextlineloc)
      tp := nextlineloc + 2
      nextlineloc := tp + strsize(tp) + 1
      texec
   else
      pauseTime := 0
      repeat c from 0 to 61
         if outputs[c>>5] & |< (c & $1F)
            pinfloat(c)
      longmove(@outputs,@outset,2)
      if s
         bytemove(tp:=@tline,s,strsize(s)+1)
      else
         putlinet(string("OK"))
         getline
      c := spaces
      if "0" =< c and c =< "9"
         insertline
         nextlineloc := eop - 2
      else
         tokenize
         if spaces
            texec

PRI processLoad : c | a
   repeat
      a := 0
      repeat
         if (c := massIn) < 0
            quit
         elseif c == _CR_ or c == _EOF_
            tline[a] := 0
            tp := @tline
            quit
         elseif c == _LF_
            next
         elseif a < linelen-1
            tline[a++] := c
      if c == _EOF_ and tline[0] == 0
         quit
      if c < 0
         abort string("Error while loading file")
      tp := @tline
      a := spaces
      if "0" =< a and a =< "9"
         insertline
         nextlineloc := eop - 2
      else
         if a <> 0
            abort string("Missing line number in file")

PRI processSave | a, c, d, p, ntoks, f0,f1,f2
   ntoks := (@tokx - @toks) / 2
   a := @progdata
   repeat while a+2 < eop
      d := wordat(a)
      massDec(d)
      massOut(" ")
      d := a + 2
      repeat while c := byte[d++]
         if c => 128
            if (c -= 128) < ntoks
               massStr(@@toks[c])
               if c <> 7   ' REM
                  massOut(" ")
            else
               massOut("{")
               massDec(c)
               massOut("}")
         else
            massOut(c)
      massOut(_CR_)
      massOut(_LF_)
      a += 3 + strsize(a+2)

PRI getAnyNumber | c, t
   case c := byte[tp]
      quote:
         if result := byte[++tp]
            if byte[++tp] == quote
              tp++
            else
               abort string("missing closing quote")
         else
            abort string("end of line in string")
      "$":
         c := byte[++tp]
         if (t := hexDigit(c)) < 0
            abort string("invalid hex character")
         result := t
         c := byte[++tp]
         repeat until (t := hexDigit(c)) < 0
            result := result << 4 | t
            c := byte[++tp]
      "%":
         c := byte[++tp]
         if not (c == "0" or c == "1")
            abort string("invalid binary character")
         result := c - "0"
         c := byte[++tp]
         repeat while c == "0" or c == "1"
            result := result << 1 | (c - "0")
            c := byte[++tp]
      "0".."9":
         result := c - "0"
         c := byte[++tp]
         repeat while c => "0" and c =< "9"
            result := result * 10 + c - "0"
            c := byte[++tp]
      other:
        abort string("invalid literal value")

PRI hexDigit(c)
'' Convert hexadecimal character to the corresponding value or -1 if invalid.
   if c => "0" and c =< "9"
      return c - "0"
   if c => "A" and c =< "F"
      return c - "A" + 10
   if c => "a" and c =< "f"
      return c - "a" + 10
   return -1

PUB dumpMemory(pin,addr,size) | i, c, p, first, buf0, buf1, buf2
'' This routine dumps a portion of the RAM/ROM to the display (pin == -1).
'' If pin is not -1, it is an EEPROM address in the form required by the
'' I2C routines in i2cSpiInit.  The specified address is or'd with this.
'' The format is 8 bytes wide with hexadecimal and ASCII.  If the address
'' is greater than $7FFFFF, then it's treated as a flash memory address
'' that starts at $800000 (translated to $000000 to $7FFFFF).  If the
'' address is greater than $FFFF (and less than $800000), it's treated as
'' an SRAM address (translated to $0000 to $FFFF).
   p := addr & $FFFFF8
   first := true
   repeat while p < (addr + size)
      if first~
         consoleHex(addr,6)
      else
         consoleHex(p,6)
      consoleOut(":")
      repeat i from 0 to 7
         byte[@buf0][i] := " "
         if p => addr and p < (addr + size)
            c := byte[p]
            consoleHex(c,2)
            if c => " " and c =< "~"
               byte[@buf0][i] := c
         else
            consoleOut(" ")
            consoleOut(" ")
         if i < 7
            consoleOut(" ")
         p++
      buf2 := 0
      consoleOut("|")
      consoleStr(@buf0)
      consoleOut("|")
      consoleOut(_CR_)

{{
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
